///////////////////////////////////////////////////////////////////////////////
/// @file   OpticalAndStereo.cpp
//
///////////////////////////////////////////////////////////////////////////////

#include "OpticalAndStereo.h"

#include <src/ImageProcessingAPI/base/Logger.h>
#include <src/ImageProcessingAPI/base/Sequence.h>

using namespace VisionRTM;
using namespace ImageProcessingAPI;

static const char* sample_spec[] =
{       "implementation_id",    "OpticalAndStereo",
        "type_name",            "OpticalAndStereo",
        "description",          "OpticalAndStereo Description",
        "version",              "1.2.1",
        "vendor",               "Systems Engineering Consultants Co., LTD.",
        "category",             "Vision RTM",
        "activity_type",        "PERIODIC",
        "kind",                 "DataFlowComponent",
        "max_instance",         "1",
        "language",             "C++",
        "lang_type",            "compile",
        "" };

OpticalAndStereo::OpticalAndStereo(RTC::Manager* manager) :
    vision_rtm::VideoFrameSyncComponentBase(manager),
    m_ParamIn("Mode",m_Mode),
    m_ResultOut("Result",m_Result)
{
    
}

OpticalAndStereo::~OpticalAndStereo()
{
}

RTC::ReturnCode_t OpticalAndStereo::onInitialize()
{
    int i, j, v, h;
    int vpitch, hpitch;
    int vs,hs;

    addInPort("Mode", m_ParamIn);
    addOutPort("Result", m_ResultOut);

    //Optical Flow
    if ((flowdata = (Flow_data_t *)malloc(sizeof(Flow_data_t) * OFNH * OFNV)) == NULL) return RTC::RTC_ERROR;
    vpitch = ((DISP_CENTER_V * 2) - (OF_VSEARCH + OF_SIZE)) / (OFNV-1);
    hpitch = ((DISP_CENTER_H * 2) - (OF_HSEARCH + OF_SIZE)) / (OFNH-1);
    vs = DISP_CENTER_V - (((OFNV - 1) * vpitch) / 2);
    hs = DISP_CENTER_H - (((OFNH - 1) * hpitch) / 2);

    for (pF = flowdata, v= vs, j = 0; j < OFNV; j++, v += vpitch) {
        for(h = hs, i = 0; i < OFNH; i++, h += hpitch, pF++) {
            pF->eh = pF->ch = h;
            pF->ev = pF->cv = v;
            pF->rh = h - CENTER_TO_REF;
            pF->rv = v - CENTER_TO_REF;
            pF->sh = pF->rh - REF_TO_SER_H;
            pF->sv = pF->rv - REF_TO_SER_V;
        }
    }

    initStereoParams();

    st0_shsize = 128;		// ステレオマッチング探索範囲水平値
    st0_svsize = 8;		// ステレオマッチング探索範囲垂直値
    st0_offset = 10;		// 探索特のオフセット横
    pitch_h = 10;		// 特徴点探索の際の水平ピッチ
    pitch_v = 10;		// 特徴点探索の際の垂直ピッチ
    c_thr = 50.0 * 8 * 8;	// コーナー特徴点の閾値
    l_thr = 60.0 * 8 * 8;	// 直線特徴点の閾値
    match_th   = 0.92; 	// マッチングの閾値

    smem_r = VRAM_ST_R(1);
    smem_l = VRAM_ST_L(1);
  
    sobelH_mem = VRAM_ST_R(8);
    sobelH_h0  = 0;
    sobelH_v0  = 0; 
    sobelV_mem = VRAM_ST_R(8);
    sobelV_h0  = 0;
    sobelV_v0  = 8; 
    
    start_h = IMAGE_HMARGIN;
    start_v = IMAGE_VMARGIN;
    size_h  = (IMG_HSIZE - 2 * IMAGE_HMARGIN) / pitch_h + 1;
    size_v  = (IMG_VSIZE - 2 * IMAGE_VMARGIN) / pitch_v + 1;
    feature_points = new SM3D_data_t2 [size_h * size_v];

    drawNum = 0;
    
    vmode = OPTICALFLOW;
    
    return RTC::RTC_OK;
}

RTC::ReturnCode_t OpticalAndStereo::onFinalize()
{
    free (flowdata);
    delete feature_points;
    return RTC::RTC_OK;
}

RTC::ReturnCode_t OpticalAndStereo::onActivated(RTC::UniqueId ec_id){

    initializeVideo(vmode);
    printf("End onActivate\n");
    return RTC::RTC_OK;
}

RTC::ReturnCode_t OpticalAndStereo::onDeactivated(RTC::UniqueId ec_id){
    RtvVideoStop();
    RtvOVLClear();
    
    return RTC::RTC_OK;
}

RTC::ReturnCode_t OpticalAndStereo::onChangeFrame(VideoMemory& captured_image)
{
    int i, j, h ,v;
    
    f0 = RtvGetFrameCount();
    RtvStartTimer();

    if (m_ParamIn.isNew())
    {
      m_ParamIn.read();
      VisionMode oldmode;

      oldmode = vmode;
      
      switch(m_Mode.data){
      case 0:
          vmode = OPTICALFLOW;
          break;
      case 1:
          vmode = STEREO;
          break;
      default:
          break;
      }
      if(vmode != oldmode) initializeVideo(vmode);
    }

    switch(vmode){
    case OPTICALFLOW:
        //OpticalFlow
        /* バーストモード開始 */
        if (RtvStartBurstCommands() != OK) {
            RtvResetBurstCommandError();
            return RTC::RTC_ERROR;
        }

        for (pF = flowdata, i = 0; i < OFNV*OFNH; i ++, pF ++) {
            if (RtvCommandSADCor8 (VRAM_Y(2), VRAM_Y(1), pF->rh, pF->rv, pF->sh, pF->sv, 0, 0, 0) !=OK){
                RtvResetBurstCommandError();
                return RTC::RTC_ERROR;
            }
            if (RtvCommandReadCorrelationMapCor8((OF_HSEARCH / 2), (OF_VSEARCH / 2), 1, 1, 0, 0) != OK) {
                RtvResetBurstCommandError();
                return RTC::RTC_ERROR;
            }
        }

        if (RtvFlushBurstCommands() != OK) {
            RtvResetBurstCommandError();
            return RTC::RTC_ERROR;
        }

        /* 結果リード */
        for (pF = flowdata, i = 0; i < OFNV*OFNH; i ++, pF ++) {
            if (RtvReturnSADCor8 (&(pF->dh), &(pF->dv), &(pF->mdist)) != OK) {
                RtvResetBurstCommandError();
                return RTC::RTC_ERROR;
            }
            if (RtvReturnReadCorrelationMapCor8(dist) != OK) {
                RtvResetBurstCommandError();
                return RTC::RTC_ERROR;
            }
    
            pF->mdistc = dist[0];
            RtvDrawLine(pF->ch, pF->cv, pF->eh, pF->ev, OVL_CLEAR);
            pF->eh = pF->ch + (pF->dh * OF_DISP_MAG);
            pF->ev = pF->cv + (pF->dv * OF_DISP_MAG);
            if ((pF->mdistc)-(pF->mdist) > 500) {
                if ((pF->dv > 0) && (pF->dv > abs(pF->dh))) {
                    dColor = OVL_RED;
                } else if ((pF->dv < 0) && (pF->dv < abs(pF->dh))) {
                    dColor = OVL_BLUE;
                } else if ((pF->dh > 0) && (pF->dh > abs(pF->dv))) {
                    dColor = OVL_YELLOW;
                } else {
                    dColor = OVL_GREEN;
                }
                RtvDrawLine(pF->ch, pF->cv, pF->eh, pF->ev, dColor);
            }
            RtvDrawDot(pF->ch, pF->cv, OVL_WHITE);
        }

        /* バーストモード停止 */
        if (RtvEndBurstCommands() != OK) {
            RtvResetBurstCommandError();
            return RTC::RTC_ERROR;
        }
    
        for(int k=0;k<OFNV*OFNH;k++){
            m_Result.data[4*k+0]=(float)flowdata[k].ch;
            m_Result.data[4*k+1]=(float)flowdata[k].cv;
            m_Result.data[4*k+2]=(float)flowdata[k].dh;
            m_Result.data[4*k+3]=(float)flowdata[k].dv;
        }

        break;
        
    case STEREO:
        /////////////////////////////////////////////////////
        // 特徴点抽出
        /////////////////////////////////////////////////////
        if (RtvStartBurstCommands() != OK) {
            RtvResetBurstCommandError();
            return RTC::RTC_ERROR;
        }
    
        RtvCommandSetSearchSize((FEATURE_SEARCH_H + FEATURE_SEARCH_H), 
                                (FEATURE_SEARCH_V + FEATURE_SEARCH_V));
        RtvCommandSetColor(S_BYTE, U_BYTE);
        RtvCommandLoadReference8(SAP_SEL_A, VRAM_Y(sobelH_mem), sobelH_h0, sobelH_v0);
        RtvCommandLoadReference8(SAP_SEL_B, VRAM_Y(sobelV_mem), sobelV_h0, sobelV_v0);
        for (i = 0, h = start_h; i < size_h; i++, h += pitch_h) {
            for (j = 0, v = start_v; j < size_v; j++, v += pitch_v) {
                if (RtvCommandDetectFeatures(VRAM_Y(smem_r), h, v) != OK) {
                    RtvResetBurstCommandError();
                    return RTC::RTC_ERROR;
                }
                if (RtvCommandGetLineFeaturePeak() != OK) {
                    RtvResetBurstCommandError();
                    return RTC::RTC_ERROR;
                }
            }
        }

        if (RtvFlushBurstCommands() != OK) {
            RtvResetBurstCommandError();
            return RTC::RTC_ERROR;
        }

        RtvReturnSetSearchSize();
        RtvReturnSetColor();
        RtvReturnLoadReference8();
        RtvReturnLoadReference8();
        fpoints = 0;
        feature = feature_points;
        for (i = 0, h = start_h; i < size_h; i++, h += pitch_h) {
            for (j = 0, v = start_v; j < size_v; j++, v += pitch_v) {
                // コーナー特徴の位置と評価値を得る
                if (RtvReturnDetectFeatures(&dh, &dv, &fval) != OK) {
                    RtvResetBurstCommandError();
                    return RTC::RTC_ERROR;
                }
                fval /= 2.0;
                if (fval > c_thr) {
                    // 特徴点の座標(固有値を計算する領域の左上座標)
                    feature->h0 = h + dh + FEATURE_SEARCH_H + FILT_OFF_H;
                    feature->v0 = v + dv + FEATURE_SEARCH_V + FILT_OFF_V;
                    feature++;
                    fpoints++;
                }
                // 線分特徴の位置と評価値を得る
                if (RtvReturnGetLineFeaturePeak (&dh, &dv, &fval) != OK) {
                    RtvResetBurstCommandError();
                    return RTC::RTC_ERROR;
                }
                fval /= 2.0;
                if (fval > l_thr) {
                    // 特徴点の座標(固有値を計算する領域の左上座標)
                    feature->h0 = h + dh + FEATURE_SEARCH_H + FILT_OFF_H;
                    feature->v0 = v + dv + FEATURE_SEARCH_V + FILT_OFF_V;
                    feature++;
                    fpoints++;
                }                
            }
        }
        
        if (RtvEndBurstCommands() != OK) {
            RtvResetBurstCommandError();
            return RTC::RTC_ERROR;
        }

        /////////////////////////////////////////////////////
        // R→Lのマッチング
        /////////////////////////////////////////////////////
        if (RtvStartBurstCommands() != OK) {
            RtvResetBurstCommandError();
            return RTC::RTC_ERROR;
        }

        // 対応点探索サイズ
        RtvCommandSetSearchSize(st0_shsize, st0_svsize);

        // 画像同士の演算なのでunsigned byte		
        RtvCommandSetColor(U_BYTE, U_BYTE); 

        for (feature = feature_points, i = 0; i < fpoints; i ++, feature ++) {
            // 探索範囲の計算:特徴点の中心座標のエピポーラ線を計算
            CalcEpiParams_L(feature->h0 + FEATURE_TEMP_CH, feature->v0 + FEATURE_TEMP_CV); 
            // 探索開始位置の計算: xは設定した最小視差,yはxとエピポーラ線から算出
            // 8x8のマッチングなので中心座標からWIN_HH,WIN_HVを引いた座標を探索開始座標として指定
            feature->sh1 = feature->h0 + st0_offset + FEATURE_TEMP_CH - WIN_HH;
            feature->sv1 = CalcEpiLv(feature->sh1 + WIN_HH) - WIN_HV - st0_svsize / 2;

            // マッチング 
            if (RtvCommandNCCCor8(VRAM_Y(smem_r), 
                                  VRAM_Y(smem_l), 
                                  feature->h0,
                                  feature->v0,
                                  feature->sh1,
                                  feature->sv1,
                                  0, 0, 0) != OK) {
                RtvResetBurstCommandError();
                return RTC::RTC_ERROR;
            }
        }

        if (RtvFlushBurstCommands() != OK) {
            RtvResetBurstCommandError();
            return RTC::RTC_ERROR;            
        }

        RtvReturnSetSearchSize();
        RtvReturnSetColor();

        for (feature = feature_points, i = 0; i < fpoints; i ++, feature ++) {
            // 最もマッチングした位置と評価値
            if (RtvReturnNCCCor8 (&match_dh, &match_dv,	&match_dist) != OK) {
                RtvResetBurstCommandError();
                return RTC::RTC_ERROR;            
            }

            // 評価値がある閾値以上で，ほかにマッチングするところがなければ
            if (match_dist > match_th) {
                feature->check = 1;
                feature->matched_h = feature->sh1 + match_dh + st0_shsize / 2;
                feature->matched_v = feature->sv1 + match_dv + st0_svsize / 2;
            } else {
                feature->check = 0;
            }
        }

        if (RtvEndBurstCommands() != OK) {
            RtvResetBurstCommandError();
            return RTC::RTC_ERROR;            
        }

        /////////////////////////////////////////////////////
        //  座標計算
        /////////////////////////////////////////////////////
        mpoints = 0;
        for (feature = feature_points, i = 0; i < fpoints; i ++, feature ++) {
            if (feature->check == 1) { 
                CalcPos(feature->h0 + WIN_HH, feature->v0 + WIN_HV,
                        feature->matched_h + WIN_HH, feature->pos);
                feature->check = 2;
                //
                m_Result.data[3 * mpoints + 0]=(float)feature->pos[0];	
                m_Result.data[3 * mpoints + 1]=(float)feature->pos[1];	
                m_Result.data[3 * mpoints + 2]=(float)feature->pos[2];	
                //
                mpoints++;
                if(feature->h0 < 380 && feature->h0 > 280 && feature->v0 > 200 && feature->v0 < 280){
                }
            }
        }
        
        /////////////////////////////////////////////////////
        //  描画
        /////////////////////////////////////////////////////
        for(i = 0; i < drawNum; i++) {
            RtvDrawBox(drawRect[i].h0, drawRect[i].v0, drawRect[i].h1, drawRect[i].v1, OVL_CLEAR);
        }

        drawNum = 0;
        //
        for (i = 0, feature = feature_points; i < fpoints; i ++, feature ++) {
            if (feature->check == 2) {
                int rh, rv;
                // ステレオマッチがどれた点の描画
                rh = (feature->h0) + WIN_HH;
                rv = (feature->v0) + WIN_HV;
                drawRect[drawNum].h0 = rh - 1;
                drawRect[drawNum].v0 = rv >> 1;
                drawRect[drawNum].h1 = rh + 3;
                drawRect[drawNum].v1 = (rv + 2) >> 1;
                RtvDrawBox(drawRect[drawNum].h0, drawRect[drawNum].v0,
                           drawRect[drawNum].h1, drawRect[drawNum].v1, OVL_BLUE);
                drawNum++;
                
                rh = (feature->matched_h) + WIN_HH;
                rv = (feature->matched_v) + WIN_HV;
                drawRect[drawNum].h0 = rh - 1;
                drawRect[drawNum].v0 = (rv >> 1) + 240;
                drawRect[drawNum].h1 = rh + 3;
                drawRect[drawNum].v1 = ((rv + 2) >> 1) + 240;
                RtvDrawBox(drawRect[drawNum].h0, drawRect[drawNum].v0,
                           drawRect[drawNum].h1, drawRect[drawNum].v1, OVL_RED);
                drawNum++;

                if (drawNum >= 8192) break;
            }
        }
        
        break;
    }

    RtvGetTimer(&time);
    RtvDrawString(16, 32, str, 3, OVL_CLEAR);
    switch(vmode){
    case OPTICALFLOW:
        sprintf(str, "t: %d", time);
        break;
    case STEREO:
        sprintf(str, "t: %d m: %d", time, mpoints);
        break;
    }
    RtvDrawString(16, 32, str, 3, OVL_GREEN);

    /* フレームオーバーフローのチェック */
    f1 = RtvGetFrameCount();
    if(f0 != f1){
        fprintf(stderr, "Warning : Frame overflow\n");
    }

    m_ResultOut.write();
    
    return RTC::RTC_OK;
}

void OpticalAndStereo::initializeVideo(VisionMode vm){
    int video_out;
    
    switch(vm){
    case OPTICALFLOW:
        video_out = VIDEO_OUTPUT_R;
        RtvOVLClear();
        RtvOVLEnable();
        RtvConfigVideoInput(1,1);
        RtvSelectVideoOutput(video_out);        

        RtvClearSearchMask(SAP_SEL_A);
        RtvClearReferenceMask8(SAP_SEL_A);

        RtvSetColor (14, 14);
        RtvSetSample(OF_SAMPLE, OF_SAMPLE, OF_SAMPLE, OF_SAMPLE);    
        RtvSetSearchSize(OF_HSEARCH, OF_VSEARCH);

        RtvSetFrameRate(1);
        RtvResetFrameCount();
        RtvVideoInput (MULTI_BUF, 3);

        m_Result.data.length(OFNH*OFNV*4);
        break;
    case STEREO:
        video_out = VIDEO_OUTPUT_BOTH;
        // 出力設定
        RtvSelectVideoOutput(video_out);
        // オーバレイ描画クリア
        RtvOVLClear();
        // オーバレイ表示オン
        RtvOVLEnable();
        // Bitmaskを全ビット1に設定
        RtvSetBitMask(0xFF);
        // カラー変換無しに設定
        RtvSetColorConversionMode(CCONV_NONE);
        // LUT変換を無効に設定
        RtvSetLUTMode(LUT_ALL_OFF);	
  
        // 探索マスククリア
        RtvClearSearchMask(SAP_SEL_A);
        RtvClearSearchMask(SAP_SEL_B);
        RtvClearSearchMask(SAP_SEL_C);
        RtvClearSearchMask(SAP_SEL_D);
        // 参照マスククリア
        RtvClearReferenceMask8(SAP_SEL_A);
        RtvClearReferenceMask8(SAP_SEL_B);
        RtvClearReferenceMask8(SAP_SEL_C);
        RtvClearReferenceMask8(SAP_SEL_D);
  
        // 参照、探索の水平、垂直の間引き設定
        RtvSetSample(0, 0, 0, 0);
  
        // 符号指定	
        RtvSetColor(S_BYTE, U_BYTE);
  
        // 空間フィルタ除数設定
        RtvSetCdiv(SAP_SEL_A, FILT_DIV);
        RtvSetCdiv(SAP_SEL_B, FILT_DIV);
        
        // 空間フィルタカーネルをメモリにライト
        RtvWriteMem(VRAM_Y(sobelH_mem), sobelH_h0, sobelH_v0, 8, 8, MaskSobelGH);
        RtvWriteMem(VRAM_Y(sobelV_mem), sobelV_h0, sobelV_v0, 8, 8, MaskSobelGV);
        
        RtvSetFrameRate(1);
        RtvResetFrameCount();
        RtvVideoInput(MULTI_BUF, 3);
        RtvWaitNextFrame();  
        
        m_Result.data.length(MAX_STEREO3D_RESULT * 3);
        break;
    }
}

void OpticalAndStereo::CalcEpiParams_L (int ru, int rv)
{
		m_LA1 = m_EP1 * ru + m_EQ1 * rv + m_ER1;
		m_LA2 = m_EP2 * ru + m_EQ2 * rv + m_ER2;
		m_LA3 = m_EP3 * ru + m_EQ3 * rv + m_ER3;
}

void OpticalAndStereo::CalcEpiParams_R (int lu, int lv)
{
		m_RA1 = lv * m_EP3 - m_EP2 - lu * m_EP1;
		m_RA2 = lv * m_ER3 - m_ER2 - lu * m_ER1;
		m_RA3 = - (lv * m_EQ3 - m_EQ2 - lu * m_EQ1);
}

int OpticalAndStereo::CalcEpiLv (int lu)
{
		int	d;

		d = (int)((m_LA1 * lu + m_LA2) / m_LA3 + 0.5);
		return(d);
}

int OpticalAndStereo::CalcEpiRv (int ru)
{
		int	d;

		d = (int)((m_RA1 * ru + m_RA2) / m_RA3 + 0.5);
		return(d);
}

void OpticalAndStereo::CalcPos (int ru, int rv, int lu, float pos[3])
{
		float	det, ir11, ir12, ir13, ir21, ir22, ir23, ir31, ir32, ir33;
		float	b1, b2, b3;
		int	luru, lurv;

		luru = lu * ru;
		lurv = lu * rv;

		det = m_Dluru * luru + m_Dlurv * lurv + m_Dlu * lu + m_Dru * ru + m_Drv * rv + m_Dc;
		ir11 = m_R11lurv * lurv + m_R11lu * lu + m_R11rv * rv + m_R11c;
		ir12 = m_R12luru * luru + m_R12lu * lu + m_R12ru * ru + m_R12c;
		ir13 = m_R13ru * ru + m_R13rv * rv + m_R13c;
		ir21 = m_R21lurv * lurv + m_R21lu * lu + m_R21rv * rv + m_R21c;
		ir22 = m_R22luru * luru + m_R22lu * lu + m_R22ru * ru + m_R22c;
		ir23 = m_R23ru * ru + m_R23rv * rv + m_R23c;
		ir31 = m_R31lurv * lurv + m_R31lu * lu + m_R31rv * rv + m_R31c;
		ir32 = m_R32luru * luru + m_R32lu * lu + m_R32ru * ru + m_R32c;
		ir33 = m_R33ru * ru + m_R33rv * rv + m_R33c;
		b1 = - (m_RP[11] * ru - m_RP[3]);
		b2 = - (m_RP[11] * rv - m_RP[7]);
		b3 = - (m_LP[11] * lu - m_LP[3]);
		pos[0] = (ir11 * b1 + ir12 * b2 + ir13 * b3) / det;
		pos[1] = (ir21 * b1 + ir22 * b2 + ir23 * b3) / det;
		pos[2] = (ir31 * b1 + ir32 * b2 + ir33 * b3) / det;
}

void OpticalAndStereo::initStereoParams (void)
{
		m_Dluru = DET_LURU; 
		m_Dlurv = DET_LURV; 
		m_Dlu = DET_LU; 
		m_Dru = DET_RU; 
		m_Drv = DET_RV; 
		m_Dc = DET_C;
		m_R11lurv = IR11_LURV; 
		m_R11lu = IR11_LU; 
		m_R11rv = IR11_RV; 
		m_R11c = IR11_C;
		m_R12luru = IR12_LURU; 
		m_R12lu = IR12_LU; 
		m_R12ru = IR12_RU; 
		m_R12c = IR12_C; 
		m_R13ru = IR13_RU; 
		m_R13rv = IR13_RV; 
		m_R13c = IR13_C;
		m_R21lurv = IR21_LURV; 
		m_R21lu = IR21_LU; 
		m_R21rv = IR21_RV; 
		m_R21c = IR21_C;
		m_R22luru = IR22_LURU; 
		m_R22lu = IR22_LU; 
		m_R22ru = IR22_RU; 
		m_R22c = IR22_C; 
		m_R23ru = IR23_RU; 
		m_R23rv = IR23_RV; 
		m_R23c = IR23_C;
		m_R31lurv = IR31_LURV; 
		m_R31lu = IR31_LU; 
		m_R31rv = IR31_RV; 
		m_R31c = IR31_C;
		m_R32luru = IR32_LURU; 
		m_R32lu = IR32_LU; 
		m_R32ru = IR32_RU; 
		m_R32c = IR32_C; 
		m_R33ru = IR33_RU; 
		m_R33rv = IR33_RV; 
		m_R33c = IR33_C;
		m_EP1 = EP1;
		m_EP2 = EP2;
		m_EP3 = EP3;
		m_EQ1 = EQ1;
		m_EQ2 = EQ2;
		m_EQ3 = EQ3;
		m_ER1 = ER1;
		m_ER2 = ER2;
		m_ER3 = ER3;
}

extern "C"
{
    void OpticalAndStereoInit(RTC::Manager* manager)
    {
        RTC::Properties profile(sample_spec);
        manager->registerFactory(profile, RTC::Create<OpticalAndStereo>,RTC::Delete<OpticalAndStereo>);
    }
}
